package org.simantics.objmap.impl;

import gnu.trove.THashMap;
import gnu.trove.TObjectIdentityHashingStrategy;

import java.util.AbstractSet;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;

import org.simantics.db.ReadGraph;
import org.simantics.db.Resource;
import org.simantics.db.WriteGraph;
import org.simantics.objmap.IFunction;
import org.simantics.objmap.ILinkType;
import org.simantics.objmap.IMapping;
import org.simantics.objmap.IMappingListener;
import org.simantics.objmap.IMappingSchema;
import org.simantics.objmap.MappingException;

/**
 * An implementation of IMapping. The class should not be created
 * directly but using methods in Mappings.
 * @see org.simantics.objmap.Mappings
 * @author Hannu Niemist
 */
public class Mapping implements IMapping {

	private static final TObjectIdentityHashingStrategy<Object> IDENTITY_HASHING =
		new TObjectIdentityHashingStrategy<Object>();
	
	IMappingSchema schema;
	
	THashMap<Resource, Link> domain = new THashMap<Resource, Link>();
	THashMap<Object, Link> range = new THashMap<Object, Link>(IDENTITY_HASHING);
	ArrayList<IMappingListener> listeners = new ArrayList<IMappingListener>();
	
	ArrayList<Link> modifiedDomainLinks = new ArrayList<Link>();
	ArrayList<Link> modifiedRangeLinks = new ArrayList<Link>();
	boolean disposed = false;
	
	boolean listensDomain; 
	
	public Mapping(IMappingSchema schema, boolean listensDomain) {
		this.schema = schema;
		this.listensDomain = listensDomain;
	}

	private void removeLink(Link link) {
		if(link.domainModified)
			modifiedDomainLinks.remove(link);
		if(link.rangeModified)
			modifiedRangeLinks.remove(link);
		link.removed = true;
	}
	
	private void createDomain(WriteGraph g, Link link) throws MappingException {
		ILinkType type = schema.linkTypeOfRangeElement(link.rangeElement);
		Resource domainElement = type.createDomainElement(g, link.rangeElement);
		
		link.type = type;
		link.domainElement = domainElement;
		domain.put(domainElement, link);
	}
	
	private void createRange(ReadGraph g, Link link) throws MappingException {
		ILinkType type = schema.linkTypeOfDomainElement(g, link.domainElement);		
		Object rangeElement = type.createRangeElement(g, link.domainElement);
		
		link.type = type;
		link.rangeElement = rangeElement;
		range.put(rangeElement, link);
	}
	
	Set<Resource> domainSet = new AbstractSet<Resource>() {

		public boolean add(Resource e) {
			if(domain.containsKey(e))
				return false;
			Link link = new Link(null, e, null);
			domain.put(e, link);
			modifiedDomainLinks.add(link);
			return true;
		}
		
		public boolean contains(Object o) {
			return domain.contains(o);
		}
		
		public boolean remove(Object o) {
			Link link = domain.remove(o);			
			if(link == null)
				return false;
			removeLink(link);
			if(link.rangeElement != null)
				range.remove(link.rangeElement);
			return true;	
		}
		
		@Override
		public Iterator<Resource> iterator() {
		    // FIXME does not implement Iterator.remove correctly
			return domain.keySet().iterator();
		}

		@Override
		public int size() {
			return domain.size();
		}
		
	};
	
	Set<Object> rangeSet = new AbstractSet<Object>() {

		public boolean add(Object e) {
			if(range.containsKey(e))
				return false;
			Link link = new Link(null, null, e);
			range.put(e, link);
			modifiedRangeLinks.add(link);
			return true;
		}
		
		public boolean contains(Object o) {
			return range.contains(o);
		}
		
		public boolean remove(Object o) {
			Link link = range.remove(o);			
			if(link == null)
				return false;
			removeLink(link);
			if(link.domainElement != null)
				domain.remove(link.domainElement);
			return true;
		}
		
		@Override
		public Iterator<Object> iterator() {
		    // FIXME does not implement Iterator.remove correctly
			return range.keySet().iterator();
		}

		@Override
		public int size() {
			return range.size();
		}
		
	};
	
	class DomainToRange implements IFunction<Resource, Object> {

		ReadGraph g;
		
		public DomainToRange(ReadGraph g) {
			this.g = g;
		}

		@Override
		public Object get(Resource element) throws MappingException {
			Link link = domain.get(element);
			if(link == null) {
				ILinkType type = schema.linkTypeOfDomainElement(g, element);
				Object rangeElement = type.createRangeElement(g, element);
				
				link = new Link(type, element, rangeElement);
				domain.put(element, link);			
				range.put(rangeElement, link);
				link.domainModified = true;
				modifiedDomainLinks.add(link);
				
				return rangeElement;				
			}
			else {
				if(link.type == null) 
					createRange(g, link);
				return link.rangeElement;
			}
		}
		
	};
	
	class RangeToDomain implements IFunction<Object, Resource> {

		WriteGraph g;
		
		public RangeToDomain(WriteGraph g) {
			this.g = g;
		}
		
		@Override
		public Resource get(Object element) throws MappingException {
			Link link = range.get(element);
			if(link == null) {
				ILinkType type = schema.linkTypeOfRangeElement(element); 
				Resource domainElement = type.createDomainElement(g, element);
				
				link = new Link(type, domainElement, element);
				domain.put(domainElement, link);			
				range.put(element, link);
				link.rangeModified = true;
				modifiedRangeLinks.add(link);
				
				return domainElement;				
			}
			else {
				if(link.type == null)
					createDomain(g, link);
				return link.domainElement;
			}
		}
		
	};
	
	@Override
	public Set<Resource> getDomain() {
		return domainSet;
	}
	
	@Override
	public Set<Object> getRange() {
		return rangeSet;
	}
	
	@Override
	public Collection<Resource> updateDomain(WriteGraph g) throws MappingException {
		RangeToDomain map = new RangeToDomain(g);
		ArrayList<Resource> updated = new ArrayList<Resource>();
		while(!modifiedRangeLinks.isEmpty()) {
			Link link = modifiedRangeLinks.remove(modifiedRangeLinks.size()-1);
			link.rangeModified = false;
			if(link.domainModified) {
				link.domainModified = false;
				modifiedDomainLinks.remove(link);
			}
			
			if(link.type == null) {
				createDomain(g, link);
			}
			
			link.type.updateDomain(g, map, link.domainElement, link.rangeElement);
			updated.add(link.domainElement);
		}		
		return updated;
	}
	
	@Override
	public Collection<Object> updateRange(ReadGraph g) throws MappingException {
		DomainToRange map = new DomainToRange(g);
		ArrayList<Object> updated = new ArrayList<Object>();
		while(!modifiedDomainLinks.isEmpty()) {
			Link link = modifiedDomainLinks.remove(modifiedDomainLinks.size()-1);
			link.domainModified = false;
			if(link.rangeModified) {
				link.rangeModified = false;
				modifiedRangeLinks.remove(link);
			}
			
			if(link.type == null) {
				createRange(g, link);
			}
			
			if(listensDomain) {
			    RangeUpdateRequest request = new RangeUpdateRequest(link, map, this);
			    g.syncRequest(request, request);
			}
			else
			    link.type.updateRange(g, map, link.domainElement, link.rangeElement);
			updated.add(link.rangeElement);
		}	
		return updated;
	}

	@Override
	public Object get(Resource domainElement) {
		Link link = domain.get(domainElement);
		if(link == null)
			return null;
		return link.rangeElement;
	}

	@Override
	public Resource inverseGet(Object rangeElement) {
		Link link = range.get(rangeElement);
		if(link == null)
			return null;
		return link.domainElement;
	}

	@Override
	public Resource inverseMap(WriteGraph g, Object rangeElement) throws MappingException {
		getRange().add(rangeElement);
		updateDomain(g);
		return inverseGet(rangeElement);
	}

	@Override
	public Object map(ReadGraph g, Resource domainElement) throws MappingException {
		getDomain().add(domainElement);
		updateRange(g);
		return get(domainElement);
	}

	void domainModified(Link link) {
	    if(!link.domainModified) {
            link.domainModified = true;
            modifiedDomainLinks.add(link);
            if(modifiedDomainLinks.size() == 1) {
                for(IMappingListener listener : listeners)
                    listener.domainModified();
            }
        }
	}
	
	@Override
	public void domainModified(Resource domainElement) {
		Link link = domain.get(domainElement);
		if(link != null)
		    domainModified(link);
	}

	void rangeModified(Link link) {
	    if(!link.rangeModified) {
            link.rangeModified = true;
            modifiedRangeLinks.add(link);
            if(modifiedRangeLinks.size() == 1) {
                for(IMappingListener listener : listeners)
                    listener.rangeModified();
            }
        }
	}
	
	@Override
	public void rangeModified(Object rangeElement) {
		Link link = range.get(rangeElement);
		if(link != null)
		    rangeModified(link);
	}

	@Override
	public boolean isDomainModified() {
		return !modifiedDomainLinks.isEmpty();
	}

	@Override
	public boolean isRangeModified() {
		return !modifiedRangeLinks.isEmpty();
	}

	@Override
	public void addMappingListener(IMappingListener listener) {
		listeners.add(listener);
	}

	@Override
	public void removeMappingListener(IMappingListener listener) {
		listeners.remove(listener);		
	}

	@Override
	public Collection<Resource> getConflictingDomainElements() {
		ArrayList<Resource> result = new ArrayList<Resource>();
		if(modifiedDomainLinks.size() < modifiedRangeLinks.size()) {
			for(Link link : modifiedDomainLinks)
				if(link.rangeModified)
					result.add(link.domainElement);
		}
		else {
			for(Link link : modifiedRangeLinks)
				if(link.domainModified)
					result.add(link.domainElement);
		}
		return result;
	}

	@Override
	public Collection<Object> getConflictingRangeElements() {
		ArrayList<Object> result = new ArrayList<Object>();
		if(modifiedDomainLinks.size() < modifiedRangeLinks.size()) {
			for(Link link : modifiedDomainLinks)
				if(link.rangeModified)
					result.add(link.rangeElement);
		}
		else {
			for(Link link : modifiedRangeLinks)
				if(link.domainModified)
					result.add(link.rangeElement);
		}
		return result;
	}

    @Override
    public void dispose() {
        disposed = true;
    }
    
    public boolean isDisposed() {
        return disposed;
    }
	
}
