package org.simantics.objmap.impl;

import gnu.trove.THashMap;
import gnu.trove.TObjectIdentityHashingStrategy;

import java.util.AbstractSet;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;

import org.simantics.db.ReadGraph;
import org.simantics.db.Resource;
import org.simantics.db.WriteGraph;
import org.simantics.objmap.IFunction;
import org.simantics.objmap.ILinkType;
import org.simantics.objmap.IMapping;
import org.simantics.objmap.IMappingListener;
import org.simantics.objmap.IMappingSchema;
import org.simantics.objmap.MappingException;

/**
 * An implementation of IMapping. The class should not be created
 * directly but using methods in Mappings.
 * @see org.simantics.objmap.Mappings
 * @author Hannu Niemist
 */
public class UnidirectionalMapping implements IMapping {

	private static final TObjectIdentityHashingStrategy<Object> IDENTITY_HASHING =
		new TObjectIdentityHashingStrategy<Object>();
	
	IMappingSchema schema;
	
	THashMap<Resource, Link> domain = new THashMap<Resource, Link>();
	ArrayList<IMappingListener> listeners = new ArrayList<IMappingListener>();
	
	ArrayList<Link> modifiedDomainLinks = new ArrayList<Link>();
	boolean disposed = false;
	
	public UnidirectionalMapping(IMappingSchema schema) {
		this.schema = schema;
	}

	private void removeLink(Link link) {
		if(link.domainModified)
			modifiedDomainLinks.remove(link);
		link.removed = true;
	}
	
	private void createDomain(WriteGraph g, Link link) throws MappingException {
		ILinkType type = schema.linkTypeOfRangeElement(link.rangeElement);
		Resource domainElement = type.createDomainElement(g, link.rangeElement);
		
		link.type = type;
		link.domainElement = domainElement;
		domain.put(domainElement, link);
	}
	
	private void createRange(ReadGraph g, Link link) throws MappingException {
		ILinkType type = schema.linkTypeOfDomainElement(g, link.domainElement);		
		Object rangeElement = type.createRangeElement(g, link.domainElement);
		
		link.type = type;
		link.rangeElement = rangeElement;
	}
	
	Set<Resource> domainSet = new AbstractSet<Resource>() {

		public boolean add(Resource e) {
			if(domain.containsKey(e))
				return false;
			Link link = new Link(null, e, null);
			domain.put(e, link);
			modifiedDomainLinks.add(link);
			return true;
		}
		
		public boolean contains(Object o) {
			return domain.contains(o);
		}
		
		public boolean remove(Object o) {
			Link link = domain.remove(o);			
			if(link == null)
				return false;
			removeLink(link);
			return true;	
		}
		
		@Override
		public Iterator<Resource> iterator() {
		    // FIXME does not implement Iterator.remove correctly
			return domain.keySet().iterator();
		}

		@Override
		public int size() {
			return domain.size();
		}
		
	};
	
	class DomainToRange implements IFunction<Resource, Object> {

		ReadGraph g;
		
		public DomainToRange(ReadGraph g) {
			this.g = g;
		}

		@Override
		public Object get(Resource element) throws MappingException {
			Link link = domain.get(element);
			if(link == null) {
				ILinkType type = schema.linkTypeOfDomainElement(g, element);
				Object rangeElement = type.createRangeElement(g, element);
				
				link = new Link(type, element, rangeElement);
				domain.put(element, link);		
				link.domainModified = true;
				modifiedDomainLinks.add(link);
				
				return rangeElement;				
			}
			else {
				if(link.type == null) 
					createRange(g, link);
				return link.rangeElement;
			}
		}
		
	};
	
	@Override
	public Set<Resource> getDomain() {
		return domainSet;
	}
	
	@Override
	public Set<Object> getRange() {
		throw new UnsupportedOperationException();
	}
	
	@Override
	public Collection<Resource> updateDomain(WriteGraph g) throws MappingException {
		throw new UnsupportedOperationException();
	}
	
	@Override
	public Collection<Object> updateRange(ReadGraph g) throws MappingException {
		DomainToRange map = new DomainToRange(g);
		ArrayList<Object> updated = new ArrayList<Object>();
		while(!modifiedDomainLinks.isEmpty()) {
			Link link = modifiedDomainLinks.remove(modifiedDomainLinks.size()-1);
			link.domainModified = false;			
			if(link.type == null) {
				createRange(g, link);
			}
			
			link.type.updateRange(g, map, link.domainElement, link.rangeElement);
			updated.add(link.rangeElement);
		}	
		return updated;
	}

	@Override
	public Object get(Resource domainElement) {
		Link link = domain.get(domainElement);
		if(link == null)
			return null;
		return link.rangeElement;
	}

	@Override
	public Resource inverseGet(Object rangeElement) {
		throw new UnsupportedOperationException();
	}

	@Override
	public Resource inverseMap(WriteGraph g, Object rangeElement) throws MappingException {
	    throw new UnsupportedOperationException();
	}

	@Override
	public Object map(ReadGraph g, Resource domainElement) throws MappingException {
		getDomain().add(domainElement);
		updateRange(g);
		return get(domainElement);
	}

	void domainModified(Link link) {
	    if(!link.domainModified) {
            link.domainModified = true;
            modifiedDomainLinks.add(link);
            if(modifiedDomainLinks.size() == 1) {
                for(IMappingListener listener : listeners)
                    listener.domainModified();
            }
        }
	}
	
	@Override
	public void domainModified(Resource domainElement) {
		Link link = domain.get(domainElement);
		if(link != null)
		    domainModified(link);
	}

	@Override
	public boolean isDomainModified() {
		return !modifiedDomainLinks.isEmpty();
	}

	@Override
	public boolean isRangeModified() {
		throw new UnsupportedOperationException();
	}

	@Override
	public void addMappingListener(IMappingListener listener) {
		listeners.add(listener);
	}

	@Override
	public void removeMappingListener(IMappingListener listener) {
		listeners.remove(listener);		
	}

	@Override
	public Collection<Resource> getConflictingDomainElements() {
	    throw new UnsupportedOperationException();
	}

	@Override
	public Collection<Object> getConflictingRangeElements() {
	    throw new UnsupportedOperationException();
	}

    @Override
    public void dispose() {
        disposed = true;
    }
    
    public boolean isDisposed() {
        return disposed;
    }

    @Override
    public void rangeModified(Object rangeElement) {
        throw new UnsupportedOperationException();
    }
	
}
