package org.simantics.objmap.schema;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;

import org.simantics.db.ReadGraph;
import org.simantics.db.exception.DatabaseException;
import org.simantics.objmap.IMappingRule;
import org.simantics.objmap.annotations.GraphType;
import org.simantics.objmap.annotations.RelatedValue;
import org.simantics.objmap.annotations.meta.HasClassRuleFactory;
import org.simantics.objmap.annotations.meta.HasFieldRuleFactory;
import org.simantics.objmap.annotations.meta.HasMethodRuleFactory;

/**
 * Static utility methods for mapping schemas
 * @author Hannu Niemist
 */
public class MappingSchemas {
    
    /**
     * Creates a new SimpleLinkType based on the annotations in the given class.
     * @throws IllegalAccessException 
     * @throws InstantiationException 
     * @see GraphType
     * @see RelatedValue
     */
	public static SimpleLinkType fromAnnotations(ReadGraph g, Class<?> clazz) throws DatabaseException, InstantiationException, IllegalAccessException {
	    GraphType graphType = clazz.getAnnotation(GraphType.class);
	    
	    ArrayList<IMappingRule> rules = new ArrayList<IMappingRule>();
	    collectRulesFromAnnotations(g, clazz, rules);
	    
	    return new SimpleLinkType(
	            g.getResourceByURI(graphType.value()), 
                clazz, rules);    
	}
	
	public static void collectRulesFromAnnotations(ReadGraph g, Class<?> clazz, Collection<IMappingRule> rules) throws DatabaseException, InstantiationException, IllegalAccessException {
        for(Annotation annotation : clazz.getAnnotations()) {
            HasClassRuleFactory factory = 
                annotation.annotationType().getAnnotation(HasClassRuleFactory.class);
            if(factory != null) {
                rules.add(factory.value().newInstance()
                        .create(g, annotation, clazz));
            }
        }

        for(Field f : clazz.getDeclaredFields()) {
            f.setAccessible(true);

            for(Annotation annotation : f.getAnnotations()) {
                HasFieldRuleFactory factory = 
                    annotation.annotationType().getAnnotation(HasFieldRuleFactory.class);
                if(factory != null) {
                    rules.add(factory.value().newInstance()
                            .create(g, annotation, f));
                }
            }
        }

        for(Method m : clazz.getDeclaredMethods()) {
            m.setAccessible(true);

            for(Annotation annotation : m.getAnnotations()) {
                HasMethodRuleFactory factory = 
                    annotation.annotationType().getAnnotation(HasMethodRuleFactory.class);
                if(factory != null) {
                    rules.add(factory.value().newInstance()
                            .create(g, annotation, m));
                }
            }
        }
    }
	
}
