package org.simantics.objmap.schema;

import java.util.ArrayList;

import org.simantics.db.ReadGraph;
import org.simantics.db.Resource;
import org.simantics.db.WriteGraph;
import org.simantics.db.exception.DatabaseException;
import org.simantics.objmap.IFunction;
import org.simantics.objmap.ILinkType;
import org.simantics.objmap.IMappingRule;
import org.simantics.objmap.MappingException;

/**
 * A link type that is associated with single domain and range type (class).
 * SimpleLinkType is composed of simpler rules whose combination determines
 * its update policy.
 * @author Hannu Niemist
 */
public class SimpleLinkType implements ILinkType {
    
    Resource domainType;
    Class<?> rangeType;
    ArrayList<IMappingRule> rules;
    
    public SimpleLinkType(Resource domainType, Class<?> rangeType,
            ArrayList<IMappingRule> rules) {
        this.domainType = domainType;
        this.rangeType = rangeType;
        this.rules = rules;
    }

    public SimpleLinkType(Resource domainType, Class<?> rangeType) {
        this(domainType, rangeType, new ArrayList<IMappingRule>());
    }

    /**
     * Adds a new rule to this link type that is enforced
     * during updates.
     */
    public void addRule(IMappingRule rule) {
        rules.add(rule);
    }
    
    @Override
    public Resource createDomainElement(WriteGraph g, Object rangeElement)
            throws MappingException {
        try {
            Resource result = g.newResource();
            g.claim(result, g.getBuiltins().InstanceOf, domainType);
            return result;
        } catch(DatabaseException e) {
            throw new MappingException(e);
        }
    }
    @Override
    public Object createRangeElement(ReadGraph g, Resource domainElement)
            throws MappingException {
        try {
            return rangeType.newInstance();
        } catch (InstantiationException e) {
            throw new MappingException(e);
        } catch (IllegalAccessException e) {
            throw new MappingException(e);
        }
    }
    @Override
    public boolean updateDomain(WriteGraph g, IFunction<Object, Resource> map,
            Resource domainElement, Object rangeElement)
            throws MappingException {
        boolean updated = false;
        for(IMappingRule rule : rules)
            rule.updateDomain(g, map, domainElement, rangeElement);
        return updated;
    }
    @Override
    public boolean updateRange(ReadGraph g, IFunction<Resource, Object> map,
            Resource domainElement, Object rangeElement)
            throws MappingException {
        boolean updated = false;
        for(IMappingRule rule : rules)
            rule.updateRange(g, map, domainElement, rangeElement);
        return updated;
    }
}
